
/* prp.js requires jsbn.js and jsbn2.js in order to function properly */

/*
 * The PRP functions presented here are based on the paper:
 * Grantham, Jon. Frobenius Pseudoprimes. Math. Comp. 70 (2001), 873-891.
 */


/* ******************************************************************
 * bnPrp: (also called a Fermat pseudoprime)
 * A "pseudoprime" to the base a is a composite number n such that,
 * (a,n)=1 and a^(n-1) = 1 mod n
 * ******************************************************************/
function bnPrp(a, disp) {
  var n = this.clone();

  if (a.compareToInt(2) < 0) return BigInteger.PRP_ERROR;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var res = n.gcd(a);
  if (res.compareToInt(1) > 0) return BigInteger.PRP_COMPOSITE;

  var nm1 = n.subtractInt(1);
  res = a.modPow(nm1, n);

  if (res.compareToInt(1) == 0) return BigInteger.PRP_PRP;
  else return BigInteger.PRP_COMPOSITE;
}/* method bnPrp */


/* *************************************************************************
 * bnEulerPrp: (also called a Solovay-Strassen pseudoprime)
 * An "Euler pseudoprime" to the base a is an odd composite number n with,
 * (a,n)=1 such that a^((n-1)/2)=(a/n) mod n [(a/n) is the Jacobi symbol]
 * *************************************************************************/
function bnEulerPrp(a, disp) {
  var n = this.clone();
  var ret = 0;

  if (a.compareToInt(2) < 0) return BigInteger.PRP_ERROR;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var res = n.gcd(a);
  if (res.compareToInt(1) > 0) return BigInteger.PRP_COMPOSITE;

  var exp = n.subtractInt(1); /* exp = n-1 */
  exp = exp.divideInt(2); /* exp = (n-1)/2 */
  res = a.modPow(exp, n);

  exp = n.addInt(a.jacobi(n));
  exp = exp.remainder(n);

  if (res.compareTo(exp) == 0) return BigInteger.PRP_PRP;
  else return BigInteger.PRP_COMPOSITE;

}/* method bnEulerPrp */


/* *********************************************************************************************
 * bnStrongPrp: (also called a Miller-Rabin pseudoprime)
 * A "strong pseudoprime" to the base a is an odd composite n = (2^r)*s+1 with s odd such that
 * either a^s == 1 mod n, or a^((2^t)*s) == -1 mod n, for some integer t, with 0 <= t < r.
 * *********************************************************************************************/
function bnStrongPrp(a, disp) {
  var n = this.clone();

  if (a.compareToInt(2) < 0) return BigInteger.PRP_ERROR;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var nm1 = n.subtractInt(1);

  /***********************************************/
  /* Find s and r satisfying: n-1=(2^r)*s, s odd */
  var r = nm1.getLowestSetBit();
  var s = nm1.shiftRight(r); /* s = nm1/2^r */

  /******************************************/
  /* Check a^((2^t)*s) mod n for 0 <= t < r */
  var y = a.modPow(s, n);
  if ( (y.compareToInt(1) == 0) || (y.compareTo(nm1) == 0) ) return BigInteger.PRP_PRP;

  while ( --r ) {
    /* y = y^2%n */
    y = y.modPowInt(2,n);
    if (y.compareTo(nm1) == 0) return BigInteger.PRP_PRP;
  }

  return BigInteger.PRP_COMPOSITE;

}/* method bnStrongPrp */


/* *******************************************************************************
 * bnLucasPrp:
 * A "Lucas pseudoprime" with parameters (P,Q) is a composite n with D=P^2-4Q,
 * (n,2QD)=1 such that U_(n-(D/n)) == 0 mod n [(D/n) is the Jacobi symbol]
 * *******************************************************************************/
function bnLucasPrp(p, q, disp) {
  var n = this.clone();
  var d = p*p - 4*q;

  if (d == 0) /* Does not produce a proper Lucas sequence */
    return BigInteger.PRP_ERROR;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var zD = nbv(d); //new BigInteger(d);

  var res = (zD.multiplyInt(2*q)).gcd(n); // res = gcd(2*q*zD, n)
  if ((res.compareTo(n) != 0) && (res.compareToInt(1) > 0))
    return BigInteger.PRP_COMPOSITE;

  /* index = n-(D/n), where (D/n) is the Jacobi symbol */
  var index = n.subtractInt(zD.jacobi(n));

  /* used for calculating the Lucas U sequence */
  var uh = nbv(1);
  var vl = nbv(2);
  var vh = nbv(p); //new BigInteger(p);
  var ql = nbv(1);
  var qh = nbv(1);
  var tmp = nbv(0);

  // s = mpz_scan1(index, 0);
  var s = index.getLowestSetBit();
  for (var j = index.bitLength()-1; j >= s+1; j--) {
    /* ql = ql*qh (mod n) */
    ql = ql.multiply(qh).remainder(n);
    if (index.testBit(j)) {
      /* qh = ql*q */
      qh = ql.multiplyInt(q);

      /* uh = uh*vh (mod n) */
      uh = uh.multiply(vh).remainder(n);

      /* vl = vh*vl - p*ql (mod n) */
      vl = vh.multiply(vl).subtract(ql.multiplyInt(p)).remainder(n);

      /* vh = vh*vh - 2*qh (mod n) */
      vh = vh.square().subtract(qh.multiplyInt(2)).remainder(n);
    }
    else
    {
      /* qh = ql */
      qh = ql.clone();

      /* uh = uh*vl - ql (mod n) */
      uh = uh.multiply(vl).subtract(ql).remainder(n);

      /* vh = vh*vl - p*ql (mod n) */
      vh = vh.multiply(vl).subtract(ql.multiplyInt(p)).remainder(n);

      /* vl = vl*vl - 2*ql (mod n) */
      vl = vl.square().subtract(ql.multiplyInt(2)).remainder(n);
    }
  }
  /* ql = ql*qh */
  ql = ql.multiply(qh);

  /* qh = ql*q */
  qh = ql.multiplyInt(q);

  /* uh = uh*vl - ql */
  uh = uh.multiply(vl).subtract(ql);

  /* vl = vh*vl - p*ql */
  vl = vh.multiply(vl).subtract(ql.multiplyInt(p));

  /* ql = ql*qh */
  ql = ql.multiply(qh);

  for (j = 1; j <= s; j++) {
    /* uh = uh*vl (mod n) */
    uh = uh.multiply(vl).remainder(n);

    /* vl = vl*vl - 2*ql (mod n) */
    vl = vl.square().subtract(ql.multiplyInt(2)).remainder(n);

    /* ql = ql*ql (mod n) */
    ql = ql.square().remainder(n);
  }

  //mpz_mod(res, uh, n); /* uh contains our return value */
  res = uh.remainder(n);

  if (res.compareToInt(0) == 0)
    return BigInteger.PRP_PRP;
  else
    return BigInteger.PRP_COMPOSITE;

}/* method bnLucasPrp */


/* *********************************************************************************************
 * bnStrongLucasPrp:
 * A "strong Lucas pseudoprime" with parameters (P,Q) is a composite n = (2^r)*s+(D/n), where
 * s is odd, D=P^2-4Q, and (n,2QD)=1 such that either U_s == 0 mod n or V_((2^t)*s) == 0 mod n
 * for some t, 0 <= t < r. [(D/n) is the Jacobi symbol]
 * *********************************************************************************************/
function bnStrongLucasPrp(p, q, disp) {
  var n = this.clone();
  var d = p*p - 4*q;

  if (d == 0) /* Does not produce a proper Lucas sequence */
    return BigInteger.PRP_ERROR;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven())
  {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var zD = nbv(d);

  var res = (zD.multiplyInt(2*q)).gcd(n); // res = gcd(2*q*zD, n)
  if ((res.compareTo(n) != 0) && (res.compareToInt(1) > 0))
    return BigInteger.PRP_COMPOSITE;

  /* nmj = n - (D/n), where (D/n) is the Jacobi symbol */
  var nmj = n.subtractInt(zD.jacobi(n));

  var r = nmj.getLowestSetBit();
  var s = nmj.shiftRight(r);

  /* make sure U_s == 0 mod n or V_((2^t)*s) == 0 mod n, for some t, 0 <= t < r */
  var uh = nbv(1);
  var vl = nbv(2);
  var vh = nbv(p);
  var ql = nbv(1);
  var qh = nbv(1);
  var tmp = nbv(0);

  for (var j = s.bitLength()-1; j >= 1; j--)
  {
    /* ql = ql*qh (mod n) */
    ql = ql.multiply(qh).remainder(n);
    if (s.testBit(j) == 1)
    {
      /* qh = ql*q */
      qh = ql.multiplyInt(q)

      /* uh = uh*vh (mod n) */
      uh = uh.multiply(vh).remainder(n);

      /* vl = vh*vl - p*ql (mod n) */
      vl = vh.multiply(vl).subtract(ql.multiplyInt(p)).remainder(n);

      /* vh = vh*vh - 2*qh (mod n) */
      vh = vh.square().subtract(qh.multiplyInt(2)).remainder(n);
    }
    else
    {
      /* qh = ql */
      qh = ql.clone();

      /* uh = uh*vl - ql (mod n) */
      uh = uh.multiply(vl).subtract(ql).remainder(n);

      /* vh = vh*vl - p*ql (mod n) */
      vh = vh.multiply(vl).subtract(ql.multiplyInt(p)).remainder(n);

      /* vl = vl*vl - 2*ql (mod n) */
      vl = vl.square().subtract(ql.multiplyInt(2)).remainder(n);
    }
  }
  /* ql = ql*qh */
  ql = ql.multiply(qh);

  /* qh = ql*q */
  qh = ql.multiplyInt(q);

  /* uh = uh*vl - ql */
  uh = uh.multiply(vl).subtract(ql);

  /* vl = vh*vl - p*ql */
  vl = vh.multiply(vl).subtract(ql.multiplyInt(p));

  /* ql = ql*qh */
  ql = ql.multiply(qh);

  uh = uh.remainder(n);
  vl = vl.remainder(n);

  /* uh contains LucasU_s and vl contains LucasV_s */
  if ((uh.compareToInt(0) == 0) || (vl.compareToInt(0) == 0))
    return BigInteger.PRP_PRP;

  for (j = 1; j < r; j++)
  {
    /* vl = vl*vl - 2*ql (mod n) */
    vl = vl.square().subtract(ql.multiplyInt(2)).remainder(n);

    /* ql = ql*ql (mod n) */
    ql = ql.square().remainder(n);

    if (vl.compareToInt(0) == 0)
      return BigInteger.PRP_PRP;
  }

  return BigInteger.PRP_COMPOSITE;

}/* method bnStrongLucasPrp */


/* ***********************************************************************************************
 * bnSelfridgePrp:
 * A "Lucas-Selfridge pseudoprime" n is a "Lucas pseudoprime" using Selfridge parameters of:
 * Find the first element D in the sequence {5, -7, 9, -11, 13, ...} such that Jacobi(D,n) = -1
 * Then use P=1 and Q=(1-D)/4 in the Lucas pseudoprime test.
 * Make sure n is not a perfect square, otherwise the search for D will only stop when D=n.
 * ***********************************************************************************************/
function bnSelfridgePrp(disp) {
  var n = this.clone();
  var d = 5, p = 1, q = 0;
  var jacobi = 0, max_d = 1000000;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var zD = nbv(d);

  /* pick p=1 and D from {5,-7,9,-11,13,...} such that jacobi(D/n)=-1, then q = (1-D)/4 */
  while (1) {
    jacobi = zD.jacobi(n);

    /* if jacobi == 0, d is a factor of n, therefore n is composite... */
    /* if d == n, then n is either prime or 9... */
    if (jacobi == 0) {
      if ((zD.abs().compareTo(n) == 0) && (zD.compareToInt(9) != 0))
        return BigInteger.PRP_PRIME;
      else
        return BigInteger.PRP_COMPOSITE;
    }
    if (jacobi == -1)
      break;

    /* if we get to the 5th d, make sure we aren't dealing with a square... */
    if (d == 13) {
      if (n.isSquare()) return BigInteger.PRP_COMPOSITE;
    }

    d = (d > 0)?(-1*(d+2)):(-1*(d-2));

    /* make sure we don't search forever */
    if (d >= max_d) return BigInteger.PRP_ERROR;

    zD = nbv(d);
  }

  q = (1-d)/4;

  return n.isLucasPrp(p, q, disp);

}/* method bnSelfridgePrp */


/* *********************************************************************************************************
 * bnStrongSelfridgePrp:
 * A "strong Lucas-Selfridge pseudoprime" n is a "strong Lucas pseudoprime" using Selfridge parameters of:
 * Find the first element D in the sequence {5, -7, 9, -11, 13, ...} such that Jacobi(D,n) = -1
 * Then use P=1 and Q=(1-D)/4 in the strong Lucase pseudoprime test.
 * Make sure n is not a perfect square, otherwise the search for D will only stop when D=n.
 * **********************************************************************************************************/
function bnStrongSelfridgePrp(disp) {
  var n = this.clone();
  var d = 5, p = 1, q = 0;
  var jacobi = 0, max_d = 1000000;

  if (n.compareToInt(2) < 0) return BigInteger.PRP_COMPOSITE;

  if (n.isEven()) {
    if (n.compareToInt(2) == 0) return BigInteger.PRP_PRIME;
    else return BigInteger.PRP_COMPOSITE;
  }

  var zD = nbv(d);

  /* pick p=1 and D from {5,-7,9,-11,13,...} such that jacobi(D/n)=-1, then q = (1-D)/4 */
  while (1) {
    jacobi = zD.jacobi(n);

    /* if jacobi == 0, d is a factor of n, therefore n is composite... */
    /* if d == n, then either n is either prime or 9... */
    if (jacobi == 0) {
      if ((zD.abs().compareTo(n) == 0) && (zD.compareToInt(9) != 0))
        return BigInteger.PRP_PRIME;
      else
        return BigInteger.PRP_COMPOSITE;
    }
    if (jacobi == -1)
      break;

    /* if we get to the 5th d, make sure we aren't dealing with a square... */
    if (d == 13) {
      if (n.isSquare()) return BigInteger.PRP_COMPOSITE;
    }

    d = (d > 0)?(-1*(d+2)):(-1*(d-2));

    /* make sure we don't search forever */
    if (d >= max_d)
      return BigInteger.PRP_ERROR;

    zD = nbv(d);
  }

  q = (1-d)/4;

  return n.isStrongLucasPrp(p, q, disp);

}/* method bnStrongSelfridgePrp */

/* **********************************************************************************
 * bnBPSW:
 * A "Baillie-Pomerance-Selfridge-Wagstaff pseudoprime" is a composite n such that
 * n is a strong pseudoprime to the base 2 and
 * n is a Lucas pseudoprime using the Selfridge parameters.
 * **********************************************************************************/
function bnBPSW(disp) {

  var ret = this.isStrongPrp(BigInteger.TWO);

  /* with a base of 2, mpz_sprp, won't return BigInteger.PRP_ERROR */
  /* so, only check for BigInteger.PRP_COMPOSITE or BigInteger.PRP_PRIME here */
  if ((ret == BigInteger.PRP_COMPOSITE) || (ret == BigInteger.PRP_PRIME))
    return ret;

  return this.isSelfridgePrp(disp);

}/* method bnBPSW */


/* ****************************************************************************************
 * bnStrongBPSW:
 * A "strong Baillie-Pomerance-Selfridge-Wagstaff pseudoprime" is a composite n such that
 * n is a strong pseudoprime to the base 2 and
 * n is a strong Lucas pseudoprime using the Selfridge parameters.
 * ****************************************************************************************/
function bnStrongBPSW(disp) {

  var ret = this.isStrongPrp(BigInteger.TWO);

  /* with a base of 2,  mpz_sprp won't return BigInteger.PRP_ERROR */
  /* so, only check for BigInteger.PRP_COMPOSITE or BigInteger.PRP_PRIME here */
  if ((ret == BigInteger.PRP_COMPOSITE) || (ret == BigInteger.PRP_PRIME))
    return ret;

  return this.isStrongSelfridgePrp(disp);

}/* method bnStrongBPSW */


// public
BigInteger.prototype.isPrp = bnPrp;
BigInteger.prototype.isStrongPrp = bnStrongPrp;
BigInteger.prototype.isEulerPrp = bnEulerPrp;
BigInteger.prototype.isLucasPrp = bnLucasPrp;
BigInteger.prototype.isStrongLucasPrp = bnStrongLucasPrp;
BigInteger.prototype.isSelfridgePrp = bnSelfridgePrp
BigInteger.prototype.isStrongSelfridgePrp = bnStrongSelfridgePrp;
BigInteger.prototype.isBPSW = bnBPSW;
BigInteger.prototype.isStrongBPSW = bnStrongBPSW;

// constants
BigInteger.PRP_ERROR = -1
BigInteger.PRP_COMPOSITE = 0
BigInteger.PRP_PRP = 1
BigInteger.PRP_PRIME = 2
